#!/usr/bin/env python

from colorama import Fore, Style
import random
import json
import requests
from typing import List, Union

from config import config
from modules.utilities.printer import printer
from modules.utilities.error_printer import error_printer


def url_opener(method: str, url: str, cookies: dict = None, data: dict = None, headers: dict = None,
               resp_type: str = 'text', name: str = '') -> List[Union[List, int, List, List, str, str]]:
    """
    This function calls sends the HTTP request based on the given method, cookies, data, headers, etc.
    and return the results in either text or JSON format. If the 'HTTP status code' is not within 200
    or 300, it tries to run it in HTTPs by replacing the scheme.

    Args:
        method (str): HTTP method request; either GET or POST.
        url (str): URL for HTTP request.
        cookies (dict, optional): Cookies for HTTP request; defaults is None.
        data (dict, optional): Data for HTTP request; defaults is None.
        headers (dict, optional): Headers for HTTP request; defaults is None.
        resp_type (str, optional): Type of response format (text/json/xml); defaults is 'text'.
        name (str, optional): Name of the request; defaults is ''.

    Returns:
        list: A list containing results with this order:
              [content, status_code, history, http_headers, version, url]
    """
    # Call the main function
    content, status_code, history, http_headers, version = requests_runner(
            method, url, cookies, data, headers, resp_type, name
        )

    # If there is any error in getting the HTTP request, try the HTTPs
    if status_code < 200 or status_code >= 400:
        url = url.replace('http://', 'https://')
        content, status_code, history, http_headers, version = requests_runner(
            method, url, cookies, data, headers, resp_type, name
        )

    # Return results
    return [
        content,
        status_code,
        history,
        http_headers,
        version,
        url  # In case the URL's scheme is changed
    ]


def requests_runner(method: str, url: str, cookies: dict = None, data: dict = None, headers: dict = None,
                    resp_type: str = 'text', name: str = '') -> List[Union[List, int, List, List, str, str]]:
    """
    This function sends the HTTP request based on the given method, cookies, data, headers, etc.
    and return the results in either text or JSON format.

    Args:
        method (str): HTTP method request; either GET or POST.
        url (str): URL for HTTP request.
        cookies (dict, optional): Cookies for HTTP request; defaults is None.
        data (dict, optional): Data for HTTP request; defaults is None.
        headers (dict, optional): Headers for HTTP request; defaults is None.
        resp_type (str, optional): Type of response format (text/json/xml); defaults is 'text'.
        name (str, optional): Name of the request; defaults is ''.

    Returns:
        list: A list containing results with this order:
              [content, status_code, history, http_headers, version, url]
    """
    # Some variables to store results
    content = []
    status_code = 0
    history = []
    http_headers = []
    version = ''

    # Print the subtitle based on the the variable "name"
    if config['verbosity'] >= 2 and name:
        printer(f'      │          ∘ {Fore.GREEN}{name} is calling{Style.RESET_ALL}')

    # Form the header and add a random User-Agent
    if not headers:
        headers = {'User-Agent': random.choice(config['user_agents'])}
    else:
        headers['User-Agent'] = random.choice(config['user_agents'])

    # Send the HTTP request based on the given method
    try:
        # Run the GET request
        if method == 'GET':
            request = requests.get(
                url,
                headers=headers,
                timeout=config['delay']['requests_timeout']
            )
        # Run the POST request
        elif method == 'POST':
            request = requests.post(
                url,
                cookies=cookies,
                data=json.dumps(data),
                headers=headers,
                timeout=config['delay']['requests_timeout']
            )

        # Get the status code
        status_code = int(request.status_code)

        # Get the history
        history = request.history

        # Get the headers
        http_headers = request.headers

        # Get the HTTP version
        version = request.raw.version

        # Get the results if there is the type is either 'text', 'xml', or 'json'
        if resp_type == 'json':
            content = json.loads(request.text)
        elif resp_type == 'text' or resp_type == 'xml':
            content = request.text
        else:
            content = request

    # region: Exceptions
    except requests.exceptions.SSLError:
        texts = [
            f'The URL {url} cannot be opened due to an SSL error.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except requests.exceptions.TooManyRedirects:
        texts = [
            f'The provided {name} URL does not seem correct.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except requests.exceptions.HTTPError:
        texts = [
            f'Error in loading the {name} URL page.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except requests.exceptions.ConnectionError:
        texts = [
            f'Error in establishing the connection to the {name} URL.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except requests.exceptions.Timeout:
        texts = [
            f'Timeout error while opening {name} URL.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except requests.exceptions.RequestException:
        texts = [
            f'Error in reading the data from the {name} URL.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except json.decoder.JSONDecodeError:
        texts = [
            f'Error in reading the JSON data retrieved from the {name} call.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except ValueError:
        texts = [
            f'No Result is found or there was an error for {name}.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    except Exception:
        texts = [
            f'An unknown error is ocurred while running {name}.',
            '',
            f'Status Code: {status_code}  -  API call URL: {url}'
        ]
        error_printer('exception', texts)
    # endregion: Exceptions

    # Return results
    return [
        content,
        status_code,
        history,
        http_headers,
        version
    ]



from urllib.parse import urlparse
from typing import List, Union

from modules.utilities.error_printer import error_printer


def url_sanitizer(url: str) -> List[Union[str, str, str]]:
    """
    this function sanitizes the given url decomposes it into all possible components
    including scheme, path, parameters, queries, fragments, etc.

    An example of a full URL:
        'http://user:pass@NetLoc:80/path;parameters/path2;parameters2?query=argument#fragment'

    All components:
        parsed_url.scheme,
        parsed_url.hostname,        # (sub)domain; it is 'netloc' in lower case
        parsed_url.netloc.lower(),  # (sub)domain
        parsed_url.path,
        parsed_url.params,
        parsed_url.query,
        parsed_url.fragment,
        parsed_url.username,
        parsed_url.password,
        parsed_url.port

    Args:
        url (str): URL string.

    Returns:
        List[str, str, str]: List of the composed full URL, the hostname, and URN.

    Raises:
        Exception: Any exception that may be raised during URL parsing.
    """
    # Strip the URL to make sure that there is no leading/tailing space
    url = url.strip()

    try:
        # Use urlparse() to split the URL into its component parts
        parsed_url = urlparse(url)

        # Check if scheme is empty, add HTTPs
        if not parsed_url.scheme:
            url = 'http://' + url
            parsed_url = urlparse(url)

    except Exception as error:
        # Handle any exceptions that may be raised
        texts = [
            f'Error in parsing the domain "{url}".',
            error
        ]
        error_printer("exception", texts)
        return ["", "", ""]

    # Return the sanitized decomposed url
    return [
        parsed_url.geturl(),
        parsed_url.hostname,
        # Replace the scheme with nothing. The last replacement is related to the case
        # that there is no scheme and 'http://' is added to it
        parsed_url.geturl().replace(parsed_url.scheme + '://', '').replace('http://', '')
    ]



#!/usr/bin/env python

from colorama import Fore, Style
import dns.resolver

from config import config
from modules.utilities.printer import printer
from modules.utilities.error_printer import error_printer

def dns_resolver(domain, record, print_args):
    """
        ### DNS Checker: DNS Dig

        This function gets the value of all DNS records.

        # Input:  - a single domain name
                - the key, subdomain, or selector.subdomain
                - a single DNS record either in an RR type (int) or case-insensitive string; e.g. 1 or A or a
        # Output: - a set of dictionaries contains a single DNS record details
    """
    results = list()

    # set the DNS server
    dns.resolver.Resolver().nameservers = config['dns']['dns_servers']

    try:
        answers = dns.resolver.resolve(domain, record)
        for answer in answers:
            results.append(answer.to_text())

    except dns.resolver.NoAnswer:
        if print_args[0]:
            ex = error_printer()
            printer(print_args[1] + Fore.RED + 'The DNS query "{0}" for "{1}" returned no answer.'.format(record, domain) + Style.RESET_ALL)
            if config['verbosity'] >= 4:
                printer(print_args[2] + 'ERROR: ' + Fore.MAGENTA + ex[0] + ' → ' + ex[4] + ':' + ex[3] + Fore.RED + '  ' + ex[1] + Style.RESET_ALL)
    except dns.resolver.NXDOMAIN:
        if print_args[0]:
            ex = error_printer()
            printer(print_args[1] + Fore.RED + 'QUERY "{0}": The domain "{1}" does not exist.'.format(record, domain) + Style.RESET_ALL)
            if config['verbosity'] >= 4:
                printer(print_args[2] + 'ERROR: ' + Fore.MAGENTA + ex[0] + ' → ' + ex[4] + ':' + ex[3] + Fore.RED + '  ' + ex[1] + Style.RESET_ALL)
    except dns.resolver.NoNameservers:
        if print_args[0]:
            ex = error_printer()
            printer(print_args[1] + Fore.RED + 'All nameservers failed to answer the DNS query "{0}" for "{1}".'.format(record, domain) + Style.RESET_ALL)
            if config['verbosity'] >= 4:
                printer(print_args[2] + 'ERROR: ' + Fore.MAGENTA + ex[0] + ' → ' + ex[4] + ':' + ex[3] + Fore.RED + '  ' + ex[1] + Style.RESET_ALL)
    except Exception:
        if print_args[0]:
            ex = error_printer()
            printer(print_args[1] + Fore.RED + 'QUERY "{0}": No result is found for "{1}" or there was an error.'.format(record, domain) + Style.RESET_ALL)
            if config['verbosity'] >= 4:
                printer(print_args[2] + 'ERROR: ' + Fore.MAGENTA + ex[0] + ' → ' + ex[4] + ':' + ex[3] + Fore.RED + '  ' + ex[1] + Style.RESET_ALL)

    # return the list of the DNS results
    finally:
        return results



#!/usr/bin/env python

from bs4 import BeautifulSoup
import re
from colorama import Fore, Back, Style
from time import sleep

from config import config
from modules.utilities.url_opener import url_opener
from modules.utilities.printer import printer


# find subdomains and related with different techniques
def subdomain_finder(domain, scan_type, existing_subdomains):
    subdomains = set()
    related_domains = set()

    # set the print arguments for the function "url_opener"
    print_args = [True, '      │        ├──■ ', '      │        │  ■■ ']

    # update the list with existing subdomains
    subdomains.update(existing_subdomains)

    # call the Hacker Target API and parse data
    def hacker_target(domain):
        subdomains = set()

        # print the subtitle: Hacker Target
        printer('      │        ├□ ' + Fore.GREEN +
              'Hacker Target API is calling' + Style.RESET_ALL)

        # download the result page of Hacker Target
        url = config['api']['hacker_target']['url_hosts'].format(domain)
        text_data = url_opener(url, '', 'text', 'Hacker Target API', print_args)[0]
        text_data = text_data.split('\n')

        # continue only if there is any data for it
        if text_data:
            # iterate over the found alternative names
            for d in text_data:
                d = d.split(',')[0]
                # ignore the main domain
                if (d != domain):
                    # find subdomains
                    if d.endswith(domain):
                        sd = d.split('.' + domain)[0]
                        subdomains.add(sd)
        return subdomains


    # call the Security Trails API and parse data
    def security_trails(domain):
        subdomains = set()

        # print the subtitle: crt.sh
        printer('      │        ├□ ' + Fore.GREEN +
              'Security Trails API is calling' + Style.RESET_ALL)

        # check if the whois API key is set
        if config['api']['security_trails']['api_key'] == '':
            printer('      │        ├──■ ' + Fore.RED +
                  'Security Trails API key is not set. Do it in the "config.py" file.' +
                  Style.RESET_ALL)
            return subdomains
        else:
            # download the result page of Security Trials
            url = config['api']['security_trails']['url_subdomain'].format(domain)
            api_key = config['api']['security_trails']['api_key']
            headers = {'APIKEY': api_key}
            result = url_opener(url, headers, 'json', 'Security Trails API', print_args)[0]
            subdomains = set(result['subdomains'])

        return subdomains


    # scrap the CRT.sh for certificates and parse them
    def crt_sh(domain, scan_type):
        subdomains = set()
        related_domains = set()
        alt_names = set()

        # get the type of the scan; quick or deep
        # if it is defined by STDIN, the setting from the config file will be ignored
        if scan_type == '':
            scan_type = config['scan_type']['ssl']

        # print the subtitle: crt.sh
        printer('      │        ├□ ' + Fore.GREEN +
              'crt.sh SSL load is running (Scan type: "{0}")'.format(scan_type) +
              Style.RESET_ALL)

        # download the certificate page on CRT
        url = config['api']['crt_sh']['url_all'].format(domain)
        json_data = url_opener(url, '', 'json', 'CRT.sh All Certs API', print_args)[0]

        # continue only if there is any data for it
        if json_data:
            json_data = sorted(json_data, key=lambda k: k['id'], reverse=True)

            # iterate over each issued certificate (cert history)
            for i in range(0, len(json_data)):
                # if it is a quick search, ignore loading each certificate
                if (scan_type == "quick") and (i > 0):
                    break
                url = config['api']['crt_sh']['url_single'].format(json_data[i]['id'])
                cert = url_opener(url, '', 'text', 'CRT.sh Single Cert API', print_args)[0]
                # fix the HTML format of the space
                cert = (cert.text).replace('&nbsp;', ' ')

                if scan_type == 'deep':
                    # print the progress
                    printer('      │      ■■■■  ' + Fore.GREEN +
                          '{0} cert(s) out of {1} certificates is loaded '.format(i+1, len(json_data)) + Fore.CYAN +
                          '({0}%)'.format(str(round((i+1)/len(json_data) * 100))) + Fore.WHITE + ' ■■■■' + Style.RESET_ALL)

                # aggregate all alternative names
                alt_names.update(re.findall(r"DNS:(.*?)<BR>", cert))

                # add a delay between downloading SSL certificates
                sleep(config['delay']['ssl'])

            # if it is a quick search
            if scan_type == 'quick':
                # iterate over the found results
                for (key, value) in enumerate(json_data):
                    # aggregate all alternative names
                    alt_names.update(value['name_value'].split('\n'))

            # iterate over the found alternative names
            for d in alt_names:
                # ignore the main domain
                if (d != domain):
                    if d.startswith('*.'):
                        d = d.split('*.')[1]
                    # find subdomains
                    if d.endswith(domain):
                        d = d.split('.' + domain)[0]
                        subdomains.add(d)
                    # find related domains (not subdomain)
                    else:
                        related_domains.add(d)

        return [subdomains, related_domains]


    # scrap the SSL Mate for certificates and parse them
    def ssl_mate(domain):
        subdomains = set()
        related_domains = set()
        alt_names = set()

        # print the subtitle: SSL Mate
        printer('      │        └□ ' + Fore.GREEN + 'SSL Mate API is running' + Style.RESET_ALL)

        # download the certificate page
        url = config['api']['ssl_mate']['url_all'].format(domain)
        json_data = url_opener(url, '', 'json', 'SSL Mate API', print_args)[0]

        # continue only if there is any data for it
        if json_data:
            # iterate over each issued certificate (cert history)
            for cert in json_data:
                # aggregate all alternative names
                alt_names.update(set(cert['dns_names']))

            # iterate over the found alternative names
            for d in alt_names:
                # ignore the main domain
                if (d != domain):
                    if d.startswith('*.'):
                        d = d.split('*.')[1]
                    # find subdomains
                    if d.endswith(domain):
                        d = d.split('.' + domain)[0]
                        subdomains.add(d)
                    # find related domains (not subdomain)
                    else:
                        related_domains.add(d)

        return [subdomains, related_domains]


    # scrap the Domain History for subdomains and parse them
    def dns_history(domain):
        subdomains = set()

        # print the subtitle: DNS History
        printer('      │        ├□ ' + Fore.GREEN + 'DNS History API is running' + Style.RESET_ALL)

        # download the DNS History page related to subdomains
        url = config['api']['dns_history']['url_subdomain'].format(domain)
        text_data = url_opener(url, '', 'text', 'DNS History API', print_args)[0]

        items = BeautifulSoup(text_data, "html.parser").find('div', {'id': 'mainarea'}).find_all('div', {'class': 'clearfix'})[1].find_all('a')

        # continue only if there is any data for it
        if items:
            # iterate over each subdomains
            for item in items:
                # aggregate all found subdomains
                subdomains.add(item.string.split('.' + domain)[0])

        return subdomains


    # get subdomains and related domains
    sub_sm, rel_sm, sub_cr, rel_cr, sub_ht ,sub_st, sub_dh = '', '', '', '', '', '', ''

    sub_ht = hacker_target(domain)
    sub_st = security_trails(domain)
    sub_dh = dns_history(domain)
    sub_cr, rel_cr = crt_sh(domain, scan_type)
    sub_sm, rel_sm = ssl_mate(domain)

    # join results
    subdomains = sub_ht.union(sub_st)
    subdomains.update(sub_cr)
    subdomains.update(sub_sm)
    subdomains.update(sub_dh)
    related_domains = rel_sm.union(rel_cr)

    # remove the domain from the set
    if subdomains and (domain in subdomains):
        subdomains.remove(domain)

    # print the title: Subdomains
    printer('      │\n      ├───' + Fore.BLACK + Back.WHITE + ' Subdomains (total: {0}) '.format(len(subdomains)) + Style.RESET_ALL)

    # print out found subdomains
    if subdomains:
        for subdomain in subdomains:
            printer('      │      ■ ' + Fore.YELLOW + subdomain + Style.RESET_ALL)
    else:
        printer('      │      ■■ ' + Fore.RED + 'No subdomain is found for "{0}".'.format(domain) + Style.RESET_ALL)

    # print the title: Related Domains
    printer('      │\n      ├───' + Fore.BLACK + Back.WHITE + ' Related Domains (total: {0}) '.format(len(related_domains)) + Style.RESET_ALL)

    # print out found related domains
    if related_domains:
        for rd in related_domains:
            printer('      │      ■ ' + Fore.YELLOW + rd + Style.RESET_ALL)
    else:
        printer('      │      ■■ ' + Fore.RED + 'No related domain is found.' + Style.RESET_ALL)

    # return the result in the format of list instead of set
    return [
        sorted(subdomains),
        sorted(related_domains)
    ]



#!/usr/bin/env python

from typing import Dict, List, Set, Union
from colorama import Fore, Style

from modules.utilities.printer import printer
from modules.utilities.url_opener import url_opener


def robots_txt(domain: str) -> Dict[str, Union[Set[str], List[Dict[str, Union[str, Set[str]]]]]]:
    """
    Fetches and parses the robots.txt file of a given domain.

    Args:
        domain (str): A string representing the domain to fetch the robots.txt file from.

    Returns:
        A dictionary with keys "sitemap" and "user_agent" (contains "name", "allow_list", and "disallow_list").
        The "sitemap" key contains a set of URLs listed in the Sitemap directive.
        The "user_agent" key contains a list of dictionaries of user-agents. The "names" contains the user-agent name,
        while "allow_list" and "disallow_list" keys contain sets of URLs allowed and disallowed by the directives.
    """
    # a variable to store cert info
    user_agents = []
    user_agent = ""
    sitemap = set()
    allow_list = set()
    disallow_list = set()
    flag = False

    # download the robots.txt
    url = f"http://{domain}/robots.txt"
    printer(f"      │        ├□ {Fore.GREEN}robots.txt is downloading{Style.RESET_ALL}")
    text_data, *_ = url_opener("GET", url, "", "", "", "text", "robots.txt file")

    # check if
    if not text_data:
        printer(f"      │        ├□ {Fore.RED}Could not download robots.txt or it does not exist{Style.RESET_ALL}")
        return {
            "sitemap": "",
            "user_agent": ""
        }

    # iterate over the lines in the robots.txt file
    for line in str(text_data).splitlines():
        # ignore comments and empty lines
        if not line or line.startswith("#"):
            continue

        # get the sitemap
        elif line.lower().startswith("sitemap:"):
            sitemap.add(line.split(":", maxsplit=1)[1].strip())

        # get the user-agent name
        elif line.lower().startswith("user-agent:"):
            # if there is a new user-agent
            # add the user-agent alongside allow/disallow list to the list
            if flag:
                user_agents.append(
                    {
                        "name": user_agent,
                        "allow_list": list(allow_list),
                        "disallow_list": list(disallow_list),
                    }
                )

                # reset variables
                sitemap = set()
                allow_list = set()
                disallow_list = set()
                user_agent = ""
                flag = False
            else:
                user_agent = line.split(":", maxsplit=1)[1].strip()
                flag = True

        # get the user-agent allow list
        elif line.startswith("Allow:"):
            allow_list.add(line.split(":", maxsplit=1)[1].strip())

        # get the user-agent disallow list
        elif line.startswith("Disallow:"):
            disallow_list.add(line.split(":", maxsplit=1)[1].strip())

    # add the last user-agent to the dictionary
    robots_txt = {
        "sitemap": list(sitemap),
        "user_agent": user_agents
    }

    return robots_txt



#!/usr/bin/env python

from colorama import Fore, Style

from config import config
from modules.utilities.url_opener import url_opener
from modules.dns.utils import dns_record_type, rr_record_finder
from modules.utilities.printer import printer


def google_dns(domain, key, record):
    """
        ### DNS: Google DNS API

        This function gets DNS records from Google DNS API.

        Read more: - https://developers.google.com/speed/public-dns/docs/doh/json
                - https://en.wikipedia.org/wiki/List_of_DNS_record_types

        # Input:  - a single domain name
                - the key, subdomain, or selector.subdomain
                - a single DNS record either in an RR type (int) or case-insensitive string; e.g. 1 or A or a
        # Output: - a set of dictionaries contains a single DNS record details
    """
    # A set to store results
    results = set()

    # Get the RR type of the DNS record
    rr_record = rr_record_finder(record)

    # Download the JSON response
    url = config['api']['google_dns']['url'].format(domain, record)
    printer('      │        ├□ ' + Fore.GREEN + 'Google DNS is downloading' + Style.RESET_ALL)
    answers = url_opener('GET', url, '', '', '', 'json', 'Google DNS API')[0]

    # Check if there is any error in the API call
    # 0 means no error
    if answers['Status'] != 0:
        printer('      │        ├■ ' + Fore.RED + f'There was an error in the API call {url}' + Style.RESET_ALL)
        return results

    # Iterate over the lines in the Google DNS API
    for answer in answers['Answer']:
        if answer['type'] != rr_record:
            printer('      │        ├■ ' + Fore.RED +
                    f'The API call {url} returned "{answer["type"]}" record instead of "{rr_record}".' +
                    Style.RESET_ALL)
            continue
        else:
            results.add({
                'record': record,
                'type': dns_record_type(rr_record, key, answer),
                'key': key,
                'value': answer['data']
            })

    return results



#!/usr/bin/env python

from typing import Dict
from colorama import Fore, Style

from modules.utilities.printer import printer
from modules.utilities.url_opener import url_opener


def security_txt(domain: str) -> Dict[str, object]:
    """
    Gets the details of the security.txt file and parses it.

    Args:
        domain: A single domain name.

    Returns:
        A dictionary contains security.txt details.
    """
    # a variable to store cert info
    security_txt = {
        'contact': set(),
        'encryption_key': set(),
        'language': set(),
        'security_policy': set(),
        'acknowledgments': set(),
        'hiring': '',
        'expiration_date': '',
        'signature_hash': '',
        'canonical': '',
    }

    # set the print arguments for the function "except_error_print"
    print_args = [True, '      │        ├──■ ', '      │        ├──■■ ']

    # download the security.txt and print the task
    printer(f'      │        ├□ {Fore.GREEN}security.txt is downloading{Style.RESET_ALL}')
    # open the page
    url = f'http://{domain}/.well-known/security.txt'
    text_data = url_opener('GET', url, '', '', '', 'text', 'security.txt file')[0]

    # check if there any 'security.txt' page exists
    if not text_data:
        printer(f'      │        ├□ {Fore.RED}Could not download security.txt or it does not exist{Style.RESET_ALL}')
        return security_txt

    # iterate over the lines in the security.txt file
    for line in str(text_data).splitlines():
        # ignore comments and empty lines
        if not line or line.startswith('#'):
            continue
        elif line.startswith('Contact:'):
            security_txt['contact'].add(line.split(':', maxsplit=1)[1].strip())
        elif line.startswith('Expires:'):
            security_txt['expiration_date'] = line.split(':', maxsplit=1)[1].strip()
        elif line.startswith('Encryption:'):
            security_txt['encryption_key'].add(line.split(':', maxsplit=1)[1].strip())
        elif line.startswith('Preferred-Languages:'):
            security_txt['language'].add(line.split(':', maxsplit=1)[1].strip().replace(' ', '').split(','))
        elif line.startswith('Canonical:'):
            security_txt['canonical'] = line.split(':', maxsplit=1)[1].strip()
        elif line.startswith('Policy:'):
            security_txt['security_policy'].add(line.split(':', maxsplit=1)[1].strip())
        elif line.startswith('Acknowledgments:'):
            security_txt['acknowledgments'].add(line.split(':', maxsplit=1)[1].strip())
        elif line.startswith('Hiring:'):
            security_txt['hiring'] = line.split(':', maxsplit=1)[1].strip()
        elif line.startswith('Hash:'):
            security_txt['signature_hash'] = line.split(':', maxsplit=1)[1].strip()

    return security_txt



#!/usr/bin/env python


"""
    ### Vulnerability: Gray Noise API

    This function gets any report provided for IPs by the Gray Noise community.

    Read more: - https://github.com/GreyNoise-Intelligence/api.greynoise.io

    # Input:  - a single IPv4 address
    # Output: - a set of dictionaries contains verdict by the Gray Noise community
"""


from config import config
from modules.utils import url_opener, error_printer
from modules.whois.utils import validate_ip


def gray_noise(ip):
    # a set to store results
    blocked_list = set()

    # continue only if the provide IP is valid and is version 4
    if validate_ip(ip) == "4":
        # download the JSON response
        url = config['api']['gray_noise']['url_ip'].format(ip)
        results = url_opener('GET', url, '', '', '', 'json', 'Gray Noise API')[0]

        # check if there is any error in the API call
        # 0 means no error
        if results['message'] != "Success":
            errors = [
                f'There was an error in the API call {url}',
                '',
                f'Message: {results["message"]}',
                '',
                ''
            ]
            error_printer(True, errors)
            return blocked_list
        else:
            blocked_list.add({
                'by': 'Gray Noise',
                'reason': results['classification'],
                'type': 'IPv4',
                'name': ip
            })

    return blocked_list



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: W3 Techs

    This function gets the list of the technologies (services) used in the given
    domain name alongside versions.

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains technologies used in the bacon
"""


import re
from time import sleep

from config import config
from modules.utils import url_opener


def w3_techs(domain):
    # variables to store results
    technologies = set()

    # get the delay defined in the config file
    delay = config['w3_techs']['delay']

    # check if report exist
    while True:
        url = config['api']['w3_techs']['url'].format(domain)
        results = url_opener('GET', url, '', '', '', 'text', 'W3 Techs')

        # find the "Crawl now!" button
        button = False if ('Crawl now!' in results) else True

        # if the doesn't exist, the report exits
        if button:
            break
        else:
            # sleep for a certain time to finish the scan
            sleep(delay)

            # send the HTTP request to scan
            data = {'add_site': '+Crawl+now!+'}
            url_opener('POST', url, '', data, '', 'text', 'W3 Techs')

    # get the results and remove data for "Site Elements" and others after it
    results = results.split('Site Elements')[0]

    # a regex to find technology and its version
    regex = r'<a href="https://w3techs.com/technologies.*?">(.*?)</a> (.*?)<br'

    # get the list of technologies and their versions
    for tech, version in re.findall(regex, results):
        # add the found technology into the set
        technologies.add(
            {
                'name': tech,
                'version': version.strip('</s>'),
                'category': '',
                'date': ''
            }
        )

    # return gathered data
    return technologies



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Security Trails API

    This function gets the list of subdomains found by Security Trails.

    Read more: https://docs.securitytrails.com/reference/domain-subdomains

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains blocked lists
"""


from datetime import datetime

from config import config
from modules.utils import url_opener, error_printer, date_formatter
from modules.subdomain.utils import sub_related_domains


def security_trails(domain):
    # variables to store results
    alt_names = set()
    subdomains = set()
    related_domains = set()

    # the HTTP headers to be sent to Security Trails
    headers = {
        "accept": "application/json",
        "APIKEY": config['api']['security_trails']['api_key']
    }

    # form the URL
    url = config['api']['security_trails']['url_subdomain'].format(domain)

    # get the results in JSON from Security Trails
    results = url_opener('GET', url, '', '', headers, 'json', 'Security Trails API')

    # if the API call returns data
    if results[1] == 200:
        # get list of the alternative names
        for alt_name in results[0]['subdomains']:
            alt_names.add(
                {
                    'name': alt_name['indicator'],
                    'reason': 'Pulse Dive',
                    'date': date_formatter(datetime.now(), '')
                }
            )

        # call the function to extract subdomain & related-domains from
        # the alternative names
        subdomains, related_domains = sub_related_domains(alt_names, domain)

    else:
        texts = [
            f'There was an error in getting data from Security Trails',
            '',
            f'Code: {results[1]} ➜ {results[0]["message"]}'
        ]
        error_printer(True, texts)

    # return gathered data
    return [
        subdomains,
        related_domains
    ]



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Censys API

    This function gets the IP address of the host and returns the list of
    services on the host (open ports) and banner grabbing.

    Read more: - https://search.censys.io/api#/hosts/viewHost

    # Input:  - a single domain name or IP address
              - the type of the asset that needs to be assessed
    # Output: - either a set contains detailed related-IPs to the given domain name
              - or a set contains details about a single IP
"""


from config import config
from modules.utils import url_opener, error_printer, json_key_checker


def Censys(bacon, type):
    # variables to store results
    ip_details = dict()
    related_ips = set()

    # for the HTTP headers to send to Censys
    headers = {
        'Authorization': f'Basic {config["api"]["censys"]["api_key"]}',
        'accept': 'application/json'
    }

    # get the results in JSON from Censys based on the type
    if type == 'domain':
        url = config['api']['censys']['url_domain'].format(bacon)
        results = url_opener('GET', url, '', '', headers, 'json', 'Censys API')[0]

        # if the API call returns data
        if results['code'] == 200:
            # get the list of IP addresses and open ports
            for details in results['result']['hits']:
                ports = set()
                technologies = set()

                # iterate over the details of each IP
                for service in details['services']:
                    ports.add(
                        {
                            'port': json_key_checker(service, ['port']),
                            'status': 'open',
                            'protocol': json_key_checker(
                                service, ['transport_protocol']
                            ),
                            'service': json_key_checker(service, ['service_name']),
                        }
                    )
                # get the list of technologies
                technologies.add(
                    {
                        'name': json_key_checker(
                            ip_details, ['operating_system', 'product']
                        ),
                        'version': '',
                    }
                )

                # form the related_ips
                related_ips.add(
                    {
                        'ip': details['ip'],
                        'ports': ports,
                        'technologies': technologies,
                        'reverse_dns': json_key_checker(
                            details, ['dns', 'reverse_dns', 'names']
                        ),
                    }
                )
        else:
            errors = [
                'There is an error in getting the data from Censys',
                '',
                f'{results["status"]} ({results["code"]}): {results["error"]}',
                '',
                ''
            ]
            error_printer(True, errors)

        # return gathered data
        return related_ips

    elif type == 'ip':
        url = config['api']['Censys']['url_ip'].format(bacon)
        results = url_opener('GET', url, '', '', headers, 'json', 'Censys API')[0]

        # if the API call returns data
        if results['code'] == 200:
            ports = set()
            technologies = set()

            # iterate over the details of the IP
            for service in details['result']['services']:
                ports.add(
                    {
                        'port': json_key_checker(service, ['port']),
                        'status': 'open',
                        'protocol': json_key_checker(service, ['transport_protocol']),
                        'service': json_key_checker(service, ['service_name']),
                    }
                )
                # get the list of technologies
                for tech in service['software']:
                    technologies.add(
                        {
                            'name': json_key_checker(tech, ['product']),
                            'version': '',
                        }
                    )

            # form the related_ips
            ip_details.add(
                {
                    'ip': details['ip'],
                    'ports': ports,
                    'technologies': technologies,
                    'reverse_dns': json_key_checker(
                        details, ['dns', 'reverse_dns', 'names']
                    ),
                    'tags': []
                }
            )
        else:
            errors = [
                'There is an error in getting the data from Censys',
                '',
                f'{results["status"]} ({results["code"]}): {results["error"]}',
                '',
                ''
            ]
            error_printer(True, errors)

        # return gathered data
        return ip_details



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Virus Total API

    This function checks if the domain name or the IP address is listed in any
    engines listed by Virus Total.

    Read more: https://developers.virustotal.com/reference

    # Input:  - a single domain name or IP address
    # Output: - a set of dictionaries contains blocked lists
"""


from datetime import datetime

from config import config
from modules.utils import url_opener, error_printer, date_formatter


def virus_total(bacon):
    # variables to store results
    blocked_list = set()

    # the HTTP headers to be sent to Virus Total
    headers = {'x-apikey': config['api']['virus_total']['api_key']}

    # form the URL based on the type of the bacon
    if type.lower() == 'ip':
        url = config['api']['virus_total']['url_ip'].format(bacon)
    if type.lower() == 'domain':
        url = config['api']['virus_total']['url_domain'].format(bacon)

    # get the results in JSON from Virus Total
    results = url_opener('GET', url, '', '', headers, 'json', 'Virus Total API')[0]

    # if the API call returns data
    if not 'error' in results:
        # get the date in epoch format
        date = ['data']['attributes']['last_analysis_date']
        date = datetime.fromtimestamp(date)

        # get the blocked lists
        results = results['data']['attributes']['last_analysis_results']

        # iterate over the results
        for result in results:
            if results['result'] in ('clean', 'unrated'):
                continue
            blocked_list.add(
                {
                    'by': result['engine_name'],
                    'reason': result['category'],
                    'type': 'IPv4',
                    'name': bacon,
                    'date': date_formatter(date , '')
                }
            )

    else:
        texts = [
            f'There was an error in getting data from Virus Total',
            '',
            f'Code: {results["code"]} ➜ {results["message"]}'
        ]
        error_printer(True, texts)

    # return gathered data
    return blocked_list



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Web Tech Survey

    This function gets the list of the technologies (services) used in the given
    domain name alongside versions and categories.

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains technologies used in the bacon
"""


import re
import json

from config import config
from modules.utils import url_opener, date_formatter


def web_tech_survey(domain):
    # variables to store results
    technologies = set()
    alt_names = set()

    # get the date format of the Rapid DNS from the config file
    date_format = config['api']['rapid_dns']['date_format']

    # get the results
    url = config['api']['web_tech_survey']['url'].format(domain)
    results = url_opener('GET', url, '', '', '', 'text', 'Web Tech Survey')

    # a regex to find technology and its version
    regex = r'<script>\s*window.__WTS_SSR_DATA__\s*=\s*(.*?)\s*</script>'
    results = json.loads(re.findall(regex, results))[0]['pageData']['results']

    # get the list of technologies and their versions
    for tech in results['AggregatedTechnologies']:
        for app in tech['Apps']:
            # add the found technology into the set
            technologies.add(
                {
                    'name': app['Technology'],
                    'version': app['Versions'][0],
                    'category': tech['Category'],
                    'date': date_formatter(tech['LastSeen'], date_format)
                }
            )
            for domain in app['Domains']:
                # add subdomains
                alt_names.add(
                    {
                        'value': domain['Domain'],
                        'reason': 'Technology',
                        'date': date_formatter(domain['LastSeen'], date_format)
                    }
                )

    # return gathered data
    return technologies



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Geek Flare

    This function gets the list of internal and external links as well as their
    HTTP status that defines which is broken and which is not

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains non-broken internal links
              - a set of dictionaries contains non-broken external links
              - a set of dictionaries contains broken internal links
              - a set of dictionaries contains broken external links
"""


from config import config
from modules.utils import url_opener, error_printer


def geek_flare_broken_link(domain):
    # variables to store results
    broken_internal_links = set()
    broken_external_links = set()
    internal_links = set()
    external_links = set()

    data = {
        "url": domain,
        "proxyCountry": config['api']['geek_flare']['proxy_country'],
        "followRedirect": True
    }
    headers = {
        'x-api-key': config['api']['geek_flare']['api_key'],
        'Content-Type': 'application/json'
    }

    # check if report exist
    url = config['api']['geek_flare']['url_broken_link']
    results = url_opener('POST', url, '', data, headers, 'json', 'Geek Flare')

    # check if there is any error
    if results['apiCode'] != 200:
        # get the list of links & specify if it is internal, external or broken
        for link in results['data']:
            if link['status'] != 404:
                if domain in link['link']:
                    # add the found broken internal link into the set
                    broken_internal_links.add(
                        {
                            'name': link['link'],
                            'status': link['status'],
                        }
                    )
                else:
                    # add the found broken external link into the set
                    broken_external_links.add(
                        {
                            'name': link['link'],
                            'status': link['status'],
                        }
                    )
            else:
                if domain in link['link']:
                    # add the found internal link into the set
                    internal_links.add(
                        {
                            'name': link['link'],
                            'status': link['status'],
                        }
                    )
                else:
                    # add the found external link into the set
                    external_links.add(
                        {
                            'name': link['link'],
                            'status': link['status'],
                        }
                    )

    else:
        errors = [
            f'There is an error in getting the data from Geek Flare',
            '',
            f'{results["message"]}',
            '',
            ''
        ]
        error_printer(True, errors)

    # return gathered data
    return [
        broken_internal_links,
        broken_external_links,
        internal_links,
        external_links
    ]



#!/usr/bin/env python

import os
import sys
import traceback


def exception_details():
    """
    This function gets details about the exceptions

    Returns:
        _description_
    """
    # get error details
    exception_type, exception_value, exception_trace = sys.exc_info()

    # get the file location
    file_name = os.path.split(exception_trace.tb_frame.f_code.co_filename)[1]

    # get the line number
    line_no = exception_trace.tb_lineno

    # get the trace back to the source
    exception_trace = ''.join(traceback.format_tb(exception_trace))

    # return results
    return [
        str(exception_type),
        str(exception_value),
        str(exception_trace),
        str(line_no),
        str(file_name)
    ]



#!/usr/bin/env python

import base64
import hashlib

from config import config


def hash_validator(source: str, hashables: set()) -> str:
    """
    This function validates a hash value of a given source and returns a term to show if it is matched
    with the hashed code in the HTML elements.

    Args:
        source (str): The source string containing the hash value.
        hashables (set): The list of elements that can be hashed

    Returns:
        str: The string defined in config file that shows if the hash value is matched or not.
    """
    # a variable to store hashed elements
    hashed_base64 = ''

    # go further only if there are elements that can be hashed
    if hashables:
        # sanitize the list of hashables
        filter(None, hashables)

        # iterate over hashable items
        for hashable in hashables:
            # a temporary variable that stores the hash algorithm
            hash_algorithm = ''

            # find the hash algorithm and the hash value
            if source.lower().startswith('sha256-'):
                sha_hashed = hashlib.sha256(hashable.encode('utf-8'))
                hash_algorithm = 'sha256'
            elif source.lower().startswith('sha384-'):
                sha_hashed = hashlib.sha384(hashable.encode('utf-8'))
                hash_algorithm = 'sha384'
            elif source.lower().startswith('sha512-'):
                sha_hashed = hashlib.sha512(hashable.encode('utf-8'))
                hash_algorithm = 'sha512'
            else:
                hashed_base64 = 'error'

            # if there is no error, get the base64 of the SHA-hash item
            if hashed_base64 != 'error':
                # get the base64 hash value of the SHA-hashed value
                hashed_base64 = base64.b64encode(sha_hashed.digest()).decode('utf8')

                # remove the shaXXX- from the source
                hashed_base64 = f'{hash_algorithm}-{hashed_base64}'

            # check if the hash is matched with the generated hash from HTML elements
            if source == hashed_base64:
                return [
                    config['html'][f'{hash_algorithm}']['matched'],
                    hashable,
                    source
                ]

        # in case hash of any hashable items doesn't match with the give hash in the source
        if hashed_base64 == 'error':
            return [
                config['html']['hash_error']['undefined'],
                hashable,
                source
            ]
        else:
            return [
                config['html'][f'{hash_algorithm}']['unmatched'],
                hashable,
                source
            ]
    # if there is not element that can be hashed while it is set in the CSP header as the source
    else:
        hash_algorithm = source.split('-')[0]

        # return unmatched as there is no element to be hashed
        return [
            config['html'][f'{hash_algorithm}']['unmatched_'],
            '',
            source,
        ]
